import { useState, useCallback, useMemo } from 'react';
import { useRouter, useMatchedRoute, useRouteChanged, useRouteMetaChanged } from './base';
import { isFunction } from '../util';

function readRouteMeta(route, key = '', props = {}) {
  let value = key && route.meta[key];

  if (isFunction(value)) {
    value = value(route, route.parent ? route.parent.children : props.router && props.router.routes, props);
  }

  return value;
}

function walkRouteTitle(router, routes = [], refresh, filter, maxLevel = 99, level = 1) {
  return routes.filter(r => r.meta.title).map(r => {
    const visible = readRouteMeta(r, 'visible', {
      router,
      level,
      maxLevel,
      refresh
    });
    if (visible === false) return;
    const title = readRouteMeta(r, 'title', {
      router,
      level,
      maxLevel,
      refresh
    });
    if (!title) return;
    if (filter && filter(r, routes, {
      title,
      visible,
      router,
      level,
      maxLevel,
      refresh
    }) === false) return;
    const ret = {
      title,
      path: r.path,
      meta: r.meta
    };

    if (level < maxLevel && Array.isArray(r.children) && r.children.length) {
      const children = walkRouteTitle(router, r.children, refresh, filter, maxLevel, level + 1);
      if (children.length) ret.children = children;
    }

    return ret;
  }).filter(Boolean);
}

function getMatched(router, currentRoute) {
  let {
    matched
  } = currentRoute;

  if (matched) {
    const isCommonPage = matched.some(r => r.meta.commonPage);

    if (isCommonPage && currentRoute.query.redirect) {
      matched = router.getMatched(currentRoute.query.redirect);
    }
  }

  return matched;
}

function getMatchedDepth(matchedRoute) {
  return matchedRoute ? matchedRoute.depth + 1 : 0;
}

function getMatchedPathList(matched, depth, maxLevel) {
  const ret = [];

  while (depth < maxLevel && matched[depth]) {
    const {
      path
    } = matched[depth++];
    ret.push(path);
  }

  return ret;
}

let dirtyTimerId;

function useRouteTitle(props = {}, defaultRouter, deps = []) {
  const {
    maxLevel = 99,
    filter,
    filterMetas = [],
    manual
  } = props;
  const router = useRouter(defaultRouter); // eslint-disable-next-line react-hooks/rules-of-hooks

  const matchedRoute = defaultRouter ? null : useMatchedRoute();
  const [dirty, setDirty] = useState(false);
  const refreshTitles = useCallback(() => setDirty(true), [setDirty]);
  const refreshTabs = useCallback((routes = []) => walkRouteTitle(router, routes, refreshTitles, filter, maxLevel), // eslint-disable-next-line react-hooks/exhaustive-deps
  [router, filter, maxLevel, refreshTitles, ...deps]);
  const [titles, setTitles] = useState(() => manual ? [] : refreshTabs(matchedRoute ? matchedRoute.config.children : router.routes));
  const [currentPaths, setCurrentPaths] = useState(() => {
    const currentRoute = router.currentRoute || router.initialRoute;
    const depth = getMatchedDepth(matchedRoute);
    return getMatchedPathList(getMatched(router, currentRoute), depth, maxLevel + depth);
  });
  useRouteChanged(router, useCallback(route => {
    const depth = getMatchedDepth(matchedRoute);
    const paths = getMatchedPathList(getMatched(router, route), depth, maxLevel + depth);

    if (paths.length !== currentPaths.length || currentPaths.some((path, i) => path !== paths[i])) {
      setCurrentPaths(paths);
    }
  }, [router, matchedRoute, maxLevel, currentPaths, setCurrentPaths]));
  const routeMetaChangedCallback = useCallback(() => setTitles(refreshTabs(matchedRoute ? matchedRoute.config.children : router.routes)), [matchedRoute, setTitles, refreshTabs, router.routes]);
  useRouteMetaChanged(router, routeMetaChangedCallback, [...filterMetas, 'title', 'visible']);
  useMemo(() => {
    if (!dirty) return;
    setDirty(false);
    if (dirtyTimerId) return;
    dirtyTimerId = setTimeout(() => {
      dirtyTimerId = null;
      router.isRunning && routeMetaChangedCallback();
    }, 0);
  }, [router, dirty, routeMetaChangedCallback]);
  return {
    titles,
    setTitles,
    refreshTitles,
    currentPaths,
    setCurrentPaths
  };
}

export { readRouteMeta, walkRouteTitle };
export default useRouteTitle;